#include "FiniteElemSpace.h"
#include "BoundaryManager.h"
#include "DOFIndexed.h"
#include "DOFVector.h"
#include "Traverse.h"
#include "BasisFunction.h"
#include "ElInfo.h"
#include "OpenMP.h"

namespace AMDiS {

  BoundaryManager::BoundaryManager(const FiniteElemSpace *feSpace)
  {
    localBounds.resize(omp_get_overall_max_threads());
    dofIndices.resize(omp_get_overall_max_threads());
    allocatedMemoryLocalBounds = feSpace->getBasisFcts()->getNumber();
    for (int i = 0; i < static_cast<int>(localBounds.size()); i++)
      localBounds[i] = new BoundaryType[allocatedMemoryLocalBounds];
  }

  BoundaryManager::BoundaryManager(BoundaryManager &bm)
  {
    localBCs = bm.localBCs;
    allocatedMemoryLocalBounds = bm.allocatedMemoryLocalBounds;
    localBounds.resize(bm.localBounds.size());
    dofIndices.resize(bm.localBounds.size());
    for (int i = 0; i < static_cast<int>(localBounds.size()); i++)
      localBounds[i] = new BoundaryType[allocatedMemoryLocalBounds];
  }

  BoundaryManager::~BoundaryManager()
  {
    for (int i = 0; i < static_cast<int>(localBounds.size()); i++)
      delete [] localBounds[i];
  }

  double BoundaryManager::boundResidual(ElInfo *elInfo, 
					DOFMatrix *matrix,
					const DOFVectorBase<double> *dv)
  {
    double result = 0.0;
    std::map<BoundaryType, BoundaryCondition*>::iterator it;
    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	result += (*it).second->boundResidual(elInfo, matrix, dv);
    
    return result;
  }

  void BoundaryManager::fillBoundaryConditions(ElInfo *elInfo, 
					       DOFVectorBase<double> *vec)
  {
    if (localBCs.size() > 0) {
      const FiniteElemSpace *feSpace = vec->getFESpace();
      Vector<DegreeOfFreedom> &dofVec = dofIndices[omp_get_thread_num()];
      const BasisFunction *basisFcts = feSpace->getBasisFcts();
      int nBasFcts = basisFcts->getNumber();
      std::map<BoundaryType, BoundaryCondition*>::iterator it;

      // get boundaries of all DOFs
      BoundaryType *localBound = localBounds[omp_get_thread_num()];
      basisFcts->getBound(elInfo, localBound);

      // get dof indices
      basisFcts->getLocalIndicesVec(elInfo->getElement(), feSpace->getAdmin(), &dofVec);

      // apply non dirichlet boundary conditions
      for (it = localBCs.begin(); it != localBCs.end(); ++it)
	if ((*it).second && !(*it).second->isDirichlet())
	  (*it).second->fillBoundaryCondition(vec, elInfo, &dofVec[0], 
					      localBound, nBasFcts);

      // apply dirichlet boundary conditions
      for (it = localBCs.begin(); it != localBCs.end(); ++it)
	if ((*it).second && (*it).second->isDirichlet())
	  (*it).second->fillBoundaryCondition(vec, elInfo, &dofVec[0], 
					      localBound, nBasFcts);
    }
  }

  void BoundaryManager::fillBoundaryConditions(ElInfo *elInfo, 
					       DOFMatrix *mat)
  {
    if (localBCs.size() <= 0)
      return;
      
    const FiniteElemSpace *feSpace = mat->getRowFESpace();
    Vector<DegreeOfFreedom> &dofVec = dofIndices[omp_get_thread_num()];
    const BasisFunction *basisFcts = feSpace->getBasisFcts();
    int nBasFcts = basisFcts->getNumber();
    std::map<BoundaryType, BoundaryCondition*>::iterator it;
    
    // get boundaries of all DOFs
    BoundaryType *localBound = localBounds[omp_get_thread_num()];
    basisFcts->getBound(elInfo, localBound);
    
    // get dof indices
    basisFcts->getLocalIndicesVec(elInfo->getElement(), 
				  feSpace->getAdmin(), &dofVec);
    
    // apply non dirichlet boundary conditions
    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if (!(*it).second->isDirichlet())
	  (*it).second->fillBoundaryCondition(mat, elInfo, &dofVec[0], 
					      localBound, nBasFcts);
    
    // apply dirichlet boundary conditions
    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if ((*it).second->isDirichlet())
	  (*it).second->fillBoundaryCondition(mat, elInfo, &dofVec[0], 
					      localBound, nBasFcts);
  }

  void BoundaryManager::initMatrix(DOFMatrix *matrix)
  {
    std::map<BoundaryType, BoundaryCondition*>::iterator it;

    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if (!(*it).second->isDirichlet())
	  (*it).second->initMatrix(matrix);

    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if ((*it).second->isDirichlet())
	  (*it).second->initMatrix(matrix);
  }

  void BoundaryManager::exitMatrix(DOFMatrix *matrix)
  {
    std::map<BoundaryType, BoundaryCondition*>::iterator it;
    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if (!(*it).second->isDirichlet())
	  (*it).second->exitMatrix(matrix);

    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if ((*it).second->isDirichlet())
	  (*it).second->exitMatrix(matrix);
  }

  void BoundaryManager::initVector(DOFVectorBase<double> *vector)
  {
    std::map<BoundaryType, BoundaryCondition*>::iterator it;
    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if (!(*it).second->isDirichlet())
	  (*it).second->initVector(vector);

    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if ((*it).second->isDirichlet())
	  (*it).second->initVector(vector);
  }

  void BoundaryManager::exitVector(DOFVectorBase<double> *vector)
  {
    std::map<BoundaryType, BoundaryCondition*>::iterator it;
    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if (!(*it).second->isDirichlet())
	  (*it).second->exitVector(vector);

    for (it = localBCs.begin(); it != localBCs.end(); ++it)
      if ((*it).second)
	if ((*it).second->isDirichlet())
	  (*it).second->exitVector(vector);
  }

}
