#include "CoarseningManager3d.h"
#include "Mesh.h"
#include "AdaptStationary.h"
#include "AdaptInstationary.h"
#include "Traverse.h"
#include "MacroElement.h"
#include "RCNeighbourList.h"
#include "FixVec.h"
#include "DOFIndexed.h"

namespace AMDiS {

  int CoarseningManager3d::coarsenFunction(ElInfo *el_info)
  {
    Tetrahedron     *el = dynamic_cast<Tetrahedron*>(const_cast<Element*>( el_info->getElement()));
    DegreeOfFreedom *edge[2];
    int             n_neigh, bound = 0;

    ElInfo *elinfo =  el_info;
    RCNeighbourList *coarsenList;


    if (el->getMark() >= 0) return 0; // el must not be coarsend, return :-(
    if (el->isLeaf()) return 0;       // single leaves don't get coarsened

    if (el->getChild(0)->getMark() >= 0  || el->getChild(1)->getMark() >= 0) {
      /****************************************************************************/
      /*  one of the children must not be coarsend; return :-(                    */
      /****************************************************************************/
      el->setMark(0);
      return 0;
    }

    if (!(el->getChild(0)->isLeaf()) || !(el->getChild(1)->isLeaf())) {
      /****************************************************************************/
      /*  one of the children is not a leaf element; try again later on           */
      /****************************************************************************/
      doMore = true;
      return 0;
    }
    /****************************************************************************/
    /*  get a list for storing all elements at the coarsening edge and fill it  */
    /****************************************************************************/
    coarsenList = new RCNeighbourList(mesh->getMaxEdgeNeigh());
    coarsenList->setCoarseningManager(this);

    /****************************************************************************/
    /*  give the refinement edge the right orientation                          */
    /****************************************************************************/

    if (el->getDOF(0,0) < el->getDOF(1,0)) {
      edge[0] = const_cast<int*>(el->getDOF(0));
      edge[1] = const_cast<int*>(el->getDOF(1));
    } else {
      edge[1] = const_cast<int*>(el->getDOF(0));
      edge[0] = const_cast<int*>(el->getDOF(1));
    }

    coarsenList->setElement(0, el, true);
    n_neigh = 1;

    coarsenList->setOppVertex(0,0,0);
    coarsenList->setElType(0, (dynamic_cast<ElInfo3d*>( el_info)->getType()));
    bound = false;
    if (getCoarsenPatch(elinfo, edge, 0, coarsenList, &n_neigh)) {
      getCoarsenPatch(elinfo, edge, 1, coarsenList, &n_neigh);
      bound = true;
    }
    coarsenList->getNeighOnPatch(n_neigh, bound);

    /****************************************************************************/
    /*  check wether we can coarsen the patch or not                            */
    /****************************************************************************/

    // ==========================================================================
    // === check for periodic boundary ==========================================
    // ==========================================================================

    if (coarsenList->doCoarsePatch(n_neigh)) {
      int n_neigh_periodic;

      DegreeOfFreedom *next_edge[2];

      RCNeighbourList *periodicList;

      while (edge[0] != NULL) {
	periodicList = coarsenList->periodicSplit(edge, 
						  next_edge,
						  &n_neigh,
						  &n_neigh_periodic);

	TEST_EXIT_DBG(periodicList)("periodicList = NULL\n");

	coarsenPatch(periodicList, n_neigh_periodic, bound);

	edge[0] = next_edge[0];
	edge[1] = next_edge[1];
      }
    }

    delete coarsenList;

    return 0;
  }

  /*****************************************************************************/
  /*  coarsenTetrahedron:  coarses a single element of the coarsening patch;   */
  /*  dofs in the interior of the element are removed and dofs in the faces of */
  /*  the element are removed if neighbour has been coarsend or if the face    */
  /*  is part of the domains boundary                                          */
  /*****************************************************************************/

  void CoarseningManager3d::coarsenTetrahedron(RCNeighbourList *coarsenList, 
					       int              index)
  {
    Tetrahedron *el = dynamic_cast<Tetrahedron*>(const_cast<Element*>( coarsenList->getElement(index))), *child[2];
    Tetrahedron *neigh;
    int      dir, el_type, i, node, opp_v;

    child[0] = dynamic_cast<Tetrahedron*>(const_cast<Element*>( el->getChild(0)));
    child[1] = dynamic_cast<Tetrahedron*>(const_cast<Element*>( el->getChild(1)));
    el_type = coarsenList->getType(index);

    /****************************************************************************/
    /*  Information about patch neighbours is still valid! But edge and face    */
    /*  dof's in a common face of patch neighbours have to be removed           */
    /****************************************************************************/

    for (dir = 0; dir < 2; dir++) {
      neigh = dynamic_cast<Tetrahedron*>(const_cast<Element*>( coarsenList->getNeighbourElement(index, dir)));
      opp_v = coarsenList->getOppVertex(index, dir);

      if (!neigh  ||  neigh->isLeaf()) {
	/****************************************************************************/
	/*  boundary face or  neigh has been coarsend: free the dof's in the face   */
	/****************************************************************************/

	if (mesh->getNumberOfDOFs(EDGE)) {
	  node = mesh->getNode(EDGE) + Tetrahedron::nChildEdge[el_type][0][dir];
	  mesh->freeDOF(const_cast<int*>( child[0]->getDOF(node)), EDGE);
	}
	if (mesh->getNumberOfDOFs(FACE))
	  {
	    node = mesh->getNode(FACE) + Tetrahedron::nChildFace[el_type][0][dir];
	    mesh->freeDOF(const_cast<int*>( child[0]->getDOF(node)), FACE);
	    node = mesh->getNode(FACE) + Tetrahedron::nChildFace[el_type][1][dir];
	    mesh->freeDOF(const_cast<int*>( child[1]->getDOF(node)), FACE);
	  }
      }
    }

    /****************************************************************************/
    /*  finally remove the interior dof's: in the common face of child[0] and   */
    /*  child[1] and at the two barycenter                                      */
    /****************************************************************************/

    if (mesh->getNumberOfDOFs(FACE)) {
      node = mesh->getNode(FACE);
      mesh->freeDOF(const_cast<int*>( child[0]->getDOF(node)), FACE);
    }


    if (mesh->getNumberOfDOFs(CENTER)) {
      node = mesh->getNode(CENTER);
      for (i = 0; i < 2; i++)
	mesh->freeDOF(const_cast<int*>( child[i]->getDOF(node)), CENTER);
    }

    /****************************************************************************/
    /*  get new data on parent and transfer data from children to parent        */
    /****************************************************************************/

    el->coarsenElementData(child[0], child[1], el_type);

    el->setFirstChild(NULL);
    el->setSecondChild(NULL);

    mesh->freeElement(child[0]);
    mesh->freeElement(child[1]);

    el->incrementMark();
  
    mesh->incrementNumberOfLeaves(-1);
    mesh->incrementNumberOfElements(-2);

    return;
  }

  /****************************************************************************/
  /*  get_coarse_patch:  gets the patch for coarsening starting on element    */
  /*  el_info->el in direction of neighbour [3-dir]; returns 1 if a boundary  */
  /*  reached and 0 if we come back to the starting element.                  */
  /*                                                                          */
  /*  if NEIGH_IN_EL we only can find the complete coarsening patch if the    */
  /*  can be coarsend; otherwise neighbour information is not valid for       */
  /*  parents; in such situation we stop looping around the edge and return 0 */
  /*                                                                          */
  /*  if !NEIGH_IN_EL we complete the loop also in the case of a incompatible */
  /*  coarsening patch since then all marks of patch elements are reset by    */
  /*  do_coarse_patch() and this minimizes calls of traverse_neighbour();     */
  /*  if we reach a boundary while looping around the edge we loop back to    */
  /*  the starting element before we return                                   */
  /****************************************************************************/
 
  bool CoarseningManager3d::getCoarsenPatch(ElInfo          *el_info, 
					    DegreeOfFreedom *edge[2],
					    int              dir, 
					    RCNeighbourList *coarsenList, 
					    int             *n_neigh)
  {
    FUNCNAME("CoarseningManager3d::getCoarsenPatch");
    ElInfo *neigh_info;
    Tetrahedron *el, *neigh;
    int         i, j, k, opp_v, edge_no;

    static unsigned char next_el[6][2] = {{3,2},
					  {1,3},
					  {1,2},
					  {0,3},
					  {0,2},
					  {0,1}};

    el = dynamic_cast<Tetrahedron*>(const_cast<Element*>( el_info->getElement()));

    if ((neigh = dynamic_cast<Tetrahedron*>(const_cast<Element*>( el_info->getNeighbour(3-dir)))) == NULL)  
      return true;

    opp_v = el_info->getOppVertex(3-dir);

    neigh_info = stack->traverseNeighbour3d(el_info, 3-dir);
    TEST_EXIT_DBG(neigh == neigh_info->getElement())
      ("neigh %d and neigh_info->el %d are not identical\n", 
       neigh->getIndex(), neigh_info->getElement()->getIndex());
    /****************************************************************************/
    /*  we have to go back to the starting element via opp_v values             */
    /*  correct information is produce by get_neigh_on_patch()                  */
    /****************************************************************************/
    coarsenList->setOppVertex(*n_neigh, 0, opp_v);  
    coarsenList->setElement(*n_neigh, neigh);
    coarsenList->setElType(*n_neigh, (dynamic_cast<ElInfo3d*>( neigh_info))->getType()); 
  
    int n_vertices = mesh->getGeo(VERTEX);
			 
    while (neigh != el) {
      for (j = 0; j < n_vertices; j++)
	if (neigh->getDOF(j) == edge[0])  break;
      for (k = 0; k < n_vertices; k++)
	if (neigh->getDOF(k) == edge[1])  break;

      if(j > 3 || k > 3) {
	for (j = 0; j < n_vertices; j++)
	  if (mesh->associated(neigh->getDOF(j, 0), edge[0][0]))  break;
	for (k = 0; k < n_vertices; k++)
	  if (mesh->associated(neigh->getDOF(k, 0), edge[1][0]))  break;
     
	TEST_EXIT_DBG(j < n_vertices  &&  k < n_vertices)
	  ("dof %d or dof %d not found on element %d with nodes (%d %d %d %d)\n", 
	   edge[0][0], edge[1][0], neigh->getIndex(), neigh->getDOF(0,0),
	   neigh->getDOF(1,0), neigh->getDOF(2,0), neigh->getDOF(3,0));
      }
      edge_no = Tetrahedron::edgeOfDOFs[j][k];
      coarsenList->setCoarsePatch(*n_neigh, edge_no == 0);

      /****************************************************************************/
      /*  get the direction of the next neighbour				    */
      /****************************************************************************/

      if (next_el[edge_no][0] != opp_v)
	i = next_el[edge_no][0];
      else
	i = next_el[edge_no][1];

      ++*n_neigh;

      opp_v = neigh_info->getOppVertex(i);
      if ((neigh = dynamic_cast<Tetrahedron*>(const_cast<Element*>( neigh_info->getNeighbour(i))))) {
	neigh_info = stack->traverseNeighbour3d(neigh_info, i);
	TEST_EXIT_DBG(neigh == neigh_info->getElement())
	  ("neigh %d and neigh_info->el %d are not identical\n", 
	   neigh->getIndex(), neigh_info->getElement()->getIndex());
	/****************************************************************************/
	/*  we have to go back to the starting element via opp_v values             */
	/*  correct information is produce by get_neigh_on_patch()                  */
	/****************************************************************************/
	coarsenList->setOppVertex(*n_neigh, 0, opp_v);  
	coarsenList->setElement(*n_neigh, neigh);
	coarsenList->setElType(*n_neigh, (dynamic_cast<ElInfo3d*>( neigh_info))->getType());
      } else
	break;
    }

    if (neigh == el)  
      return false;

    /****************************************************************************/
    /*  the domain's boundary is reached; loop back to the starting el          */
    /****************************************************************************/

    i = *n_neigh-1;
    opp_v = coarsenList->getOppVertex(i, 0);
    do {
      TEST_EXIT_DBG(neigh_info->getNeighbour(opp_v)  &&  i > 0)
	("while looping back domains boundary was reached or i == 0\n");
      opp_v = coarsenList->getOppVertex(i--, 0);
      neigh_info = stack->traverseNeighbour3d(neigh_info, opp_v);
    } while(neigh_info->getElement() != el);

    return true;
  }

  /****************************************************************************/
  /*  coarse_patch: first rebuild the dofs on the parents then do restriction */
  /*  of data (if possible) and finally coarsen the patch elements            */
  /****************************************************************************/

  void CoarseningManager3d::coarsenPatch(RCNeighbourList *coarsenList, 
					 int              n_neigh, 
					 int              bound)
  {
    Tetrahedron *el = dynamic_cast<Tetrahedron*>(const_cast<Element*>( coarsenList->getElement(0)));
    int      i, node;
    DegreeOfFreedom *dof;

    if (mesh->getNumberOfDOFs(EDGE)) {
      /****************************************************************************/
      /*  get dof for coarsening edge                                             */
      /****************************************************************************/
      node = mesh->getNode(EDGE);
      if (!(dof = const_cast<int*>( el->getDOF(node))))
	dof = mesh->getDOF(EDGE);
    } else
      dof = NULL;

    if (mesh->getNumberOfDOFs(EDGE) || 
	mesh->getNumberOfDOFs(FACE) || 
	mesh->getNumberOfDOFs(CENTER))
      {
	for (i = 0; i < n_neigh; i++)
	  coarsenList->addDOFParent(i, dof);
      }

    /****************************************************************************/
    /*  restrict dof vectors to the parents on the patch                        */
    /****************************************************************************/

    int iadmin;
    int nrAdmin = mesh->getNumberOfDOFAdmin();
    for(iadmin = 0; iadmin < nrAdmin; iadmin++) {
      std::list<DOFIndexedBase*>::iterator it;
      DOFAdmin* admin = const_cast<DOFAdmin*>(&mesh->getDOFAdmin(iadmin));
      std::list<DOFIndexedBase*>::iterator end = admin->endDOFIndexed();
      for(it = admin->beginDOFIndexed(); it != end; ++it)
	(*it)->coarseRestrict(*coarsenList, n_neigh);
    } 

    /****************************************************************************/
    /*  and now start to coarsen the patch:                                     */
    /****************************************************************************/
    /*  remove dof's of the coarsening edge                                     */
    /****************************************************************************/

    mesh->freeDOF(const_cast<int*>( el->getChild(0)->getDOF(3)), VERTEX);
    mesh->incrementNumberOfVertices(-1);

    if (mesh->getNumberOfDOFs(EDGE))
      {
	node = mesh->getNode(EDGE)+2;
	mesh->freeDOF(const_cast<int*>( el->getChild(0)->getDOF(node)), EDGE);
	mesh->freeDOF(const_cast<int*>( el->getChild(1)->getDOF(node)), EDGE);
      }

    if (coarsenList->getElement(0)->isNewCoordSet()) {
      coarsenList->getElement(0)->eraseNewCoord();
    }

    for (i = 0; i < n_neigh; i++)
      {
	coarsenList->getElement(i)->setNewCoord(NULL);
	coarsenTetrahedron(coarsenList, i);
      }

    /****************************************************************************/
    /*  if the coarsening edge is an interior edge there are  n_neigh + 1 edges */
    /*  and 2*n_neigh + 1 faces removed; if it is a boundary edge it is one     */
    /*  more edge and one more face                                             */
    /****************************************************************************/

    if (bound) {
      mesh->incrementNumberOfEdges(-(n_neigh + 2));
      mesh->incrementNumberOfFaces(-(2*n_neigh + 1));
    } else {
      mesh->incrementNumberOfEdges(-(n_neigh + 1));
      mesh->incrementNumberOfFaces(-(2*n_neigh));
    }

    return;
  }

}
