// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ElInfo.h */

#ifndef AMDIS_ELINFO_H
#define AMDIS_ELINFO_H

#include <boost/numeric/mtl/mtl.hpp>

#include "Flag.h"
#include "Boundary.h"
#include "Global.h"
#include "FixVec.h"
#include "Element.h"
#include "AMDiS_fwd.h" 

namespace AMDiS {

  /** \ingroup Traverse
   * \brief 
   * An ElInfo object holds informations wich are not stored in the corresponding
   * element. It is filled during mesh traversal by the traversal routines.
   * A fill flag determines which informations should be filled and which elements
   * should be visited. Since it is a
   * pure virtual base class for the dimension speciefic ElInfo classes, it must
   * not be instantiated directly.
   * \see ElInfo1d \see ElInfo2d \see ElInfo3d
   */

  class ElInfo
  {
  protected: 
    /// Protected constructor. Avoids instatiation of the basis class
    ElInfo();

    /** \brief
     * Protected constructor. Avoids instatiation of the basis class.
     * \param mesh pointer to the corresponding mesh.
     */
    ElInfo(Mesh *mesh);

  public:
    /// Virtual destructor because ElInfo is pure virtual.
    virtual ~ElInfo();

    /** \brief
     * Assignement operator.
     * \param rhs right hand side.
     */
    ElInfo& operator=(const ElInfo& rhs) 
    {
      mesh_ = rhs.mesh_;
      element_ = rhs.element_;
      parent_ = rhs.parent_;
      macroElement_ = rhs.macroElement_;
      fillFlag_ = rhs.fillFlag_;
      level = rhs.level;
      iChild = rhs.iChild;
      coord_ = rhs.coord_;
      boundary_ = rhs.boundary_;
      oppCoord_ = rhs.oppCoord_;
      neighbour_ = rhs.neighbour_;
      neighbourCoord_ = rhs.neighbourCoord_;
      oppVertex_ = rhs.oppVertex_;
      return *this;
    }

  public:
    /** \name getting methods
     * \{ 
     */

    /// Get ElInfo's \ref mesh_
    inline Mesh* getMesh() const 
    { 
      return mesh_; 
    }

    /// Get ElInfo's \ref macroElement_
    inline MacroElement* getMacroElement() const 
    { 
      return macroElement_; 
    }

    /// Get ElInfo's \ref element
    inline Element* getElement() const 
    { 
      return element_; 
    }

    /// Get ElInfo's \ref parent_
    inline Element* getParent() const 
    { 
      return parent_; 
    }

    /// Get ElInfo's \ref fillFlag_
    inline Flag getFillFlag() const 
    { 
      return fillFlag_; 
    }

    /// Get ElInfo's \ref level
    inline int getLevel() const 
    { 
      return level; 
    }

    /// Get ElInfo's \ref iChild
    inline int getIChild() const 
    {
      return iChild;
    }

    /** \brief
     * Get ElInfo's \ref coord_[i]. This is a WorldVector<double> filled with the
     * coordinates of the i-th vertex of element \ref el.
     */
    inline WorldVector<double>& getCoord(int i) 
    { 
      return coord_[i]; 
    }

    /** \brief
     * Get ElInfo's \ref coord_[i]. This is a WorldVector<double> filled with the
     * coordinates of the i-th vertex of element \ref el.
     */
    inline const WorldVector<double>& getCoord(int i) const 
    { 
      return coord_[i]; 
    }

    /** \brief
     * Get ElInfo's \ref coord_. This is a FixVec<WorldVector<double> > filled with the
     * coordinates of the all vertice of element \ref el.
     */
    inline FixVec<WorldVector<double>, VERTEX>& getCoords() 
    { 
      return coord_; 
    }

    /** \brief
     * Get ElInfo's \ref coord_. This is a FixVec<WorldVector<double> > filled with the
     * coordinates of the all vertice of element \ref el.
     */
    inline const FixVec<WorldVector<double>, VERTEX>& getCoords() const 
    { 
      return coord_; 
    }

    /// Get ElInfo's \ref oppCoord_[i]
    inline WorldVector<double>& getOppCoord(int i) 
    { 
      return oppCoord_[i]; 
    }

    /// Get ElInfo's \ref boundary_[i] 
    inline BoundaryType getBoundary(int i) const 
    { 
      return boundary_[i]; 
    }

    /// Get boundary type of i-th vertex/edge/face (pos).
    BoundaryType getBoundary(GeoIndex pos, int i);

    /// Get ElInfo's \ref neighbour_[i]
    inline Element* getNeighbour(int i) const 
    { 
      return neighbour_[i]; 
    }

    /// Get ElInfo's \ref neighbourCoord_[i]
    inline FixVec<WorldVector<double>, VERTEX> getNeighbourCoord(int i) const 
    {
      return neighbourCoord_[i];
    }

    /// Get ElInfo's \ref oppVertex_[i] 
    inline unsigned char getOppVertex(int i) const 
    { 
      return oppVertex_[i]; 
    }

    virtual int getSideOfNeighbour(int i) 
    { 
      return oppVertex_[i]; 
    }

    /// Get ElInfo's \ref det_
    inline double getDet() const 
    { 
      return det_; 
    }

    /// Returns \ref grdLambda
    inline const DimVec<WorldVector<double> >& getGrdLambda() const 
    { 
      return grdLambda; 
    }

    /// Returns \ref projection_[i]
    inline Projection *getProjection(int i) const 
    {
      return projection_[i];
    }

    /// Returns \ref parametric_
    inline bool getParametric() 
    { 
      return parametric_; 
    }

    virtual void getSubElemCoordsMat(mtl::dense2D<double>& mat,
				     int basisFctsDegree) const {}

    virtual void getSubElemCoordsMat_so(mtl::dense2D<double>& mat,
					int basisFctsDegree) const {}

    /** \} */ 

    /** \name setting methods
     * \{ 
     */

    /// Set ElInfo's \ref mesh_
    inline void setMesh(Mesh* aMesh) 
    { 
      mesh_ = aMesh; 
    }

    /// Set ElInfo's \ref macroElement_
    inline void setMacroElement(MacroElement* mel) 
    { 
      macroElement_ = mel; 
    }

    /// Set ElInfo's \ref element
    inline void setElement(Element* elem) 
    { 
      element_ = elem; 
    }

    /// Set ElInfo's \ref parent_
    inline void setParent(Element* elem) 
    { 
      parent_ = elem; 
    }

    /// Set ElInfo's \ref fillFlag_
    inline void setFillFlag(Flag flag) 
    { 
      fillFlag_ = flag; 
    }

    /// Sets ElInfo's \ref coord_[i]. 
    inline void setCoord(int i,WorldVector<double>& coord) 
    { 
      coord_[i] = coord; 
    }

    /// Sets ElInfo's \ref coord. 
    inline void setCoords(FixVec<WorldVector<double>,VERTEX >& coords) 
    { 
      coord_ = coords; 
    }

    /// Set ElInfo's \ref level
    inline void setLevel(int l) 
    { 
      level = l; 
    }

    /// Set ElInfo's \ref boundary_[i] 
    inline void setBoundary(int i, BoundaryType t) 
    { 
      boundary_[i] = newBound(boundary_[i], t);
    }

    /// Set \ref projection_[i] = p
    inline void setProjection(int i, Projection *p) 
    {
      projection_[i] = p;
    }

    /// Set \ref det_ = d
    inline void setDet(double d) 
    { 
      det_ = d; 
    }

    /// Set \ref parametric_ = param
    inline void setParametric(bool param) 
    { 
      parametric_ = param; 
    }

    /// Set \ref refinementPath
    inline void setRefinementPath(unsigned long rPath)
    {
      refinementPath = rPath;
    } 

    /// Set \ref refinementPathLength
    inline void setRefinementPathLength(int length)
    {
      refinementPathLength = length;
    }

    /** \} */


    /** \brief
     * Returns the absolute value of the determinant of the affine linear 
     * parametrization's Jacobian
     */
    virtual double calcDet() const;

    /** \brief
     * Used by non static method \ref calcDet(). Calculates the determinant
     * for a given vector of vertex coordinates.
     */
    double calcDet(const FixVec<WorldVector<double>, VERTEX> &coords) const;

    /** \brief
     * Checks whether flag is set in ElInfo's \ref fillFlag_. If not, the program
     * exits.
     */
    void testFlag(const Flag& flag) const;

    /** \brief
     * Transforms local barycentric coordinates of a point defined on this element
     * to global world coordinates.
     */
    void coordToWorld(const DimVec<double>& lambda,
		      WorldVector<double>& world) const;
  
    /// Fills ElInfo's \ref det_ and \ref grdLambda entries.
    virtual void fillDetGrdLambda();

    /** \brief
     * Returns a pointer to a vector, which contains the barycentric coordinates
     * with respect to \ref element of a point with world coordinates world.
     * The barycentric coordinates are stored in lambda. 
     * pure virtual => must be overriden in sub-class.
     */
    virtual const int worldToCoord(const WorldVector<double>& world, 
				   DimVec<double>* lambda) const = 0;

    /** \brief
     * Fills this ElInfo with macro element information of mel.
     * pure virtual => must be overriden in sub-class.
     */
    virtual void fillMacroInfo(const MacroElement *mel) = 0;

    /** \brief
     * Fills this ElInfo for the child ichild using hierarchy information and
     * parent data parentInfo.
     * pure virtual => must be overriden in sub-class.
     */
    virtual void fillElInfo(int ichild, const ElInfo *parentInfo) = 0;

    /** \brief
     * calculates the Jacobian of the barycentric coordinates on \element and stores
     * the matrix in grd_lam. The return value of the function is the absolute
     * value of the determinant of the affine linear paraetrization's Jacobian.
     * pure virtual => must be overriden in sub-class.
     */
    virtual double calcGrdLambda(DimVec<WorldVector<double> >& grd_lam) = 0;

    /** \brief
     * calculates a normal of the given side (1d,2d: edge, 3d: face) of \ref element.
     * Returns the absolute value of the determinant of the
     * transformation to the reference element.
     * pure virtual => must be overriden in sub-class.
     */
    virtual double getNormal(int side, WorldVector<double> &normal) = 0;


    /** \brief
     * calculates a normal of the element in dim of world = dim + 1.
     * Returns the absolute value of the determinant of the
     * transformation to the reference element.
     * pure virtual => must be overriden in sub-class.
     */
    virtual double getElementNormal(WorldVector<double> &elementNormal) const 
    {
      FUNCNAME("ElInfo::getElementNormal()");

      ERROR("virtual function not implemented in this sub-class ");
    
      return(0.0);
    }

  protected:
    /// Pointer to the current mesh
    Mesh *mesh_;

    /// Pointer to the current element
    Element *element_;

    /// \ref element is child of element parent_
    Element *parent_;

    /** \brief 
     * \ref element is an element of the binary tree located at MacroElement 
     * macroElement_
     */
    MacroElement *macroElement_;

    /** \brief
     * Indicates wich elements will be called and wich information should be
     * present while mesh traversal.
     */
    Flag fillFlag_;

    /** \brief 
     * Level of the element. The level is zero for macro elements and the level
     * of the children is (level of the parent + 1). level_ is filled always by
     * the traversal routines.
     */
    unsigned char level;

    /** \brief
     * This ElInfo is the iChild-th child of the parent element.
     */
    int iChild;

    /** \brief 
     * \ref coord_[i] is a WorldVector<double> storing the world coordinates of the
     * i-th vertex of element \ref element.
     */
    FixVec<WorldVector<double>, VERTEX> coord_;

    /** \brief 
     * boundary_[i] is the BoundaryType of the i-th edge/face
     * for i=0,...,N_NEIGH - 1. In 3d 
     * (*boundary_)[N_FACES + i] is a pointer to the Boundary
     * object of the i-th edge, for i=0,..,N_EDGES - 1. It is
     * a pointer to NULL for an interior edge/face.
     */
    FixVec<BoundaryType, BOUNDARY> boundary_;

    /** \brief
     * Vector storing pointers to projections for each face, edge, vertex.
     */
    FixVec<Projection*, PROJECTION> projection_;

    /** \brief 
     * oppCoord_[i] coordinates of the i-th neighbour vertex opposite the
     * common edge/face.
     */
    FixVec<WorldVector<double>, NEIGH> oppCoord_;

    /** \brief 
     * neighbour_[i] pointer to the element at the edge/face with local index i.
     * It is a pointer to NULL for boundary edges/faces.
     */
    FixVec<Element*, NEIGH> neighbour_;

    /** \brief
     * neighbourCoord_[i][j] are the coordinate of the j-th vertex of the i-th
     * neighbour element with the common edge/face.
     */
    FixVec<FixVec<WorldVector<double>, VERTEX>, NEIGH> neighbourCoord_;

    /** \brief 
     * oppVertex_[i] is undefined if neighbour_[i] is a pointer to NULL. 
     * Otherwise it is the local index of the neighbour's vertex opposite the
     * common edge/face.
     */
    FixVec<unsigned char, NEIGH> oppVertex_;

    /// Elements determinant.
    double det_;

    /// Gradient of lambda.
    DimVec<WorldVector<double> > grdLambda;

    /// True, if this elInfo stores parametrized information. False, otherwise.
    bool parametric_;

    /// Stores the world dimension.
    int dimOfWorld;

    unsigned long refinementPath;

    int refinementPathLength;

    /** \brief
     * This is a transformation matrix used during dual traverse. It is set, if
     * the current element is the smaller element of an element pair in the traverse.
     * Then this matrix defines a mapping for points defined in barycentric 
     * coordinates on the larger element, to the barycentric coordinates of the smaller
     * element.
     */
    mtl::dense2D<double> subElemCoordsMat;

    mtl::dense2D<double> subElemCoordsMat_so;

  public:
    /** \brief 
     * child_vertex[el_type][child][i] = father's local vertex index of new 
     * vertex i. 4 stands for the newly generated vertex .
     */
    static const int childVertex[3][2][4];

    /** \brief 
     * child_edge[el_type][child][i] = father's local edge index of new edge i.
     * new edge 2 is half of old edge 0, new edges 4,5 are really new edges, and
     * value is different: child_edge[][][4,5] = index of same edge in other
     * child.
     */
    static const int childEdge[3][2][6];

    friend class ElInfo1d;
    friend class ElInfo2d;
    friend class ElInfo3d;
  };

}

#include "ElInfo1d.h"
#include "ElInfo2d.h"
#include "ElInfo3d.h"

#endif  // AMDIS_ELINFO_H








