// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ITL_Preconditioner.h */

#ifndef AMDIS_ITL_SOLVER_H
#define AMDIS_ITL_SOLVER_H

#include "ITL_OEMSolver.h"
#include <boost/numeric/itl/krylov/cg.hpp>
#include <boost/numeric/itl/krylov/bicg.hpp>
#include <boost/numeric/itl/krylov/bicgstab.hpp>
#include <boost/numeric/itl/krylov/bicgstab_2.hpp>
#include <boost/numeric/itl/krylov/bicgstab_ell.hpp>

namespace AMDiS {

  // ============================================================================
  // ===== class CGSolver =======================================================
  // ============================================================================
  
  /// Helper class to establish a type from a function
  class cg_solver_type
  {
  public:
    template < class LO, class X, class B, class L, class R, class I >
    int operator()(const LO& A, X& x, const B& b, const L& l, const R& r, I& iter)
    { return itl::cg(A, x, b, l, r, iter); }
  };


  /**
   * \ingroup Solver
   * 
   * \brief
   * Solves a linear system by the conjugate gradient method and can be used for
   * symmetric positive definite system matrices.
   */
  class CGSolver : public ITL_OEMSolver<cg_solver_type>
  {
  public:
    /// The constructor reads required parameters and sets solvers \ref name.
    CGSolver(std::string name) : ITL_OEMSolver<cg_solver_type>(name) {}
  };


  // ============================================================================
  // ===== class BiCGSolver =====================================================
  // ============================================================================

  /// Helper class to establish a type from a function
  class bicg_solver_type
  {
  public:
    template < class LO, class X, class B, class L, class R, class I >
    int operator()(const LO& A, X& x, const B& b, const L& l, const R&, I& iter)
    { return itl::bicg(A, x, b, l, iter); }
  };

  /**  
   * \ingroup Solver
   *
   * \brief
   * Solves a linear system by a stabilized BiCG method and can be used for 
   * system matrices.
   */
  class BiCGSolver : public ITL_OEMSolver<bicg_solver_type>
  {
  public:
    /// The constructor reads required parameters and sets solvers \ref name.
    BiCGSolver(std::string name) : ITL_OEMSolver<bicg_solver_type>(name) {}
  };


  // ============================================================================
  // ===== class BiCGStab =======================================================
  // ============================================================================

  /// Helper class to establish a type from a function
  class bicgstab_type
  {
  public:
    template < class LinOp, class X, class B, class L, class R, class I >
    int operator()(const LinOp& A, X& x, const B& b, const L& l, const R&, I& iter)
    { return itl::bicgstab(A, x, b, l, iter); }
  };


  /**  
   * \ingroup Solver
   *
   * \brief
   * Solves a linear system by a stabilized BiCG method and can be used for 
   * system matrices.
   */
  class BiCGStab : public ITL_OEMSolver<bicgstab_type>
  {
  public:
    /// The constructor reads required parameters and sets solvers \ref name.
    BiCGStab(std::string name) : ITL_OEMSolver<bicgstab_type>(name) {}
  };


  // ============================================================================
  // ===== class BiCGStab2 ======================================================
  // ============================================================================

  /// Helper class to establish a type from a function
  class bicgstab2_type
  {
  public:
    template < class LinOp, class X, class B, class L, class R, class I >
    int operator()(const LinOp& A, X& x, const B& b, const L& l, const R&, I& iter)
    { return itl::bicgstab_2(A, x, b, l, iter); }
  };


  /**  
   * \ingroup Solver
   *
   * \brief
   * Solves a linear system by a stabilized BiCG method and can be used for 
   * system matrices.
   */
  class BiCGStab2 : public ITL_OEMSolver<bicgstab2_type>
  {
  public:
    /// The constructor reads required parameters and sets solvers \ref name.
    BiCGStab2(std::string name) : ITL_OEMSolver<bicgstab2_type>(name) {}
  };


  // ============================================================================
  // ===== class BiCGStabEll ======================================================
  // ============================================================================

  /// Helper class to establish a type from a function
  class bicgstab_ell_type
  {
  public:
    template < class LinOp, class X, class B, class L, class R, class I >
    int operator()(const LinOp& A, X& x, const B& b, const L& l, const R& r, I& iter, int ell)
    { return itl::bicgstab_ell(A, x, b, l, r, iter, ell); }
  };

  /**  
   * \ingroup Solver
   *
   * \brief
   * Solves a linear system by a stabilized BiCG method and can be used for 
   * system matrices.
   */
  class BiCGStabEll : public ITL_OEMSolver_para<bicgstab_ell_type>
  {
  public:
    /// The constructor reads required parameters and sets solvers \ref name.
    BiCGStabEll(std::string name) : ITL_OEMSolver_para<bicgstab_ell_type>(name) {}
  };


} // namespace AMDiS

#endif // AMDIS_ITL_SOLVER
 
