#include "NonLinUpdater.h"
#include "Global.h"
#include "DOFMatrix.h"
#include "DOFVector.h"
#include "Traverse.h"
#include "ElInfo.h"
#include "Flag.h"
#include "Operator.h"
#include "FiniteElemSpace.h"
#include "BasisFunction.h"
#include "BoundaryManager.h"

namespace AMDiS {

  void NonLinUpdaterScal::update(bool updateDF,
				 DOFVector<double> *F)
  {
    FUNCNAME("NonLinUpdaterScal::update()");

    TraverseStack stack;
    ElInfo *el_info;
    Flag fill_flag;

    TEST_EXIT(F || df)("neither F nor df are set\n");

    const FiniteElemSpace *feSpace = 
      F ? F->getFESpace() : df->getFESpace();

    if (updateDF)
      df->clear();

    BasisFunction *basFcts = const_cast<BasisFunction*>(feSpace->getBasisFcts());

    if (F) {
      F->set(0.0);
    }

    fill_flag = 
      Mesh::CALL_LEAF_EL|
      Mesh::FILL_COORDS|
      Mesh::FILL_BOUND|
      Mesh::FILL_DET|
      Mesh::FILL_GRD_LAMBDA;

    BoundaryType *bound = new BoundaryType[feSpace->getBasisFcts()->getNumber()];

    el_info = stack.traverseFirst(feSpace->getMesh(), -1, fill_flag);
    while (el_info) {
      basFcts->getBound(el_info, bound);

      if (updateDF)
	df->assemble(1.0, el_info, bound);
      if (F)
	F->assemble(1.0, el_info, bound);
    
      el_info = stack.traverseNext(el_info);
    }

    delete [] bound;
  }

  void NonLinUpdaterVec::update(bool updateDF,
				SystemVector *F)
  {
    FUNCNAME("NonLinUpdaterVec::update()");

    TraverseStack stack;
    ElInfo *el_info;
    Flag fill_flag;

    TEST_EXIT(F || df)("neither F nor df are set\n");

    if (!updateDF && (F == NULL)) {
      WARNING("No RHS vector and no update for system matrix! Updater does nothing!\n");
      return;
    }

    int size = df->getNumRows();

    const FiniteElemSpace *feSpace = 
      F ? 
      F->getDOFVector(0)->getFESpace() : 
      (*df)[0][0]->getFESpace();

    if (updateDF) {
      TEST_EXIT(df)("df not set but update tried!\n");

      for (int i = 0; i < size; i++)
	for (int j = 0; j < size; j++)
	  if ((*df)[i][j])
	    (*df)[i][j]->clear();
    }

    BasisFunction *basFcts = const_cast<BasisFunction*>(feSpace->getBasisFcts());

    if (F) {
      for (int i = 0; i < size; i++) {
	F->getDOFVector(i)->set(0.0);
      }
    }

    fill_flag = 
      Mesh::CALL_LEAF_EL|
      Mesh::FILL_COORDS|
      Mesh::FILL_BOUND|
      Mesh::FILL_DET|
      Mesh::FILL_GRD_LAMBDA;

    BoundaryType *bound = new BoundaryType[basFcts->getNumber()];

    el_info = stack.traverseFirst(feSpace->getMesh(), -1, fill_flag);
    while (el_info) {
      basFcts->getBound(el_info, bound);

      if (updateDF)
	for (int i = 0; i < size; i++)
	  for (int j = 0; j < size; j++)
	    if ((*df)[i][j])
	      (*df)[i][j]->assemble(1.0, el_info, bound);

      if (F)
	for(int i = 0; i < size; i++)
	  F->getDOFVector(i)->assemble(1.0, el_info, bound);
    
      el_info = stack.traverseNext(el_info);
    }

    delete [] bound;
  }

}
