// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file PardisoSolver.h */

#ifndef AMDIS_PARDISOSOLVER_H
#define AMDIS_PARDISOSOLVER_H

#ifdef HAVE_MKL

#include <mkl.h>

#include "OEMSolver.h"

#if defined(MKL_ILP64)
#define MKL_INT long long
#else
#define MKL_INT int
#endif

namespace AMDiS {

  /**
   * \ingroup Solver
   * 
   * \brief
   * Wrapper for the external Intel MKL Pardiso solver
   *
   * This is a direct solver for large sparse matrices.
   */
  template<typename VectorType>
  class PardisoSolver : public OEMSolver<VectorType>
  {
  public:
    /** \brief
     * Creator class used in the OEMSolverMap.
     */
    class Creator : public OEMSolverCreator<VectorType>
    {
    public:
      virtual ~Creator() {};
      
      /** \brief
       * Returns a new UmfPackSolver object.
       */
      OEMSolver<VectorType>* create() { 
	return new PardisoSolver<VectorType>(this->name); 
      };
    };

    /** \brief
     * constructor
     */
    PardisoSolver(std::string name);
    
    /** \brief
     * destructor
     */
    ~PardisoSolver();
  protected:
    /** \brief
     * Implements OEMSolver<VectorType>::init().
     */
    void init() {
      p = this->vectorCreator->create();
      r = this->vectorCreator->create();
    };
    
    /** \brief
     * Implements OEMSolver<VectorType>::exit().
     */
    void exit() {
      this->vectorCreator->free(p);
      this->vectorCreator->free(r);
    };
    
    /** \brief
     * Implements OEMSolver<VectorType>::solve().
     */
    int solveSystem(MatVecMultiplier<VectorType> *mv, 
		    VectorType *x, VectorType *b, bool reuseMatrix);
    
  private:
    /** \brief
     * These vectors are justed to calculate the final residual of the solution.
     */
    VectorType *r, *p;
  };
  
}

#include "PardisoSolver.hh"

#endif // HAVE_MKL

#endif // AMDIS_PARDISOSOLVER_H
