// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file RecoveryEstimator.h */

#ifndef AMDIS_MYRECOVERYESTIMATOR_H
#define AMDIS_MYRECOVERYESTIMATOR_H

#include "Estimator.h"
#include "Recovery.h"

namespace AMDiS {

  /// Error estimator using the recovery gradient of the finite element solution.
  class RecoveryEstimator : public Estimator
  {
  public:
    /// Creator class.
    class Creator : public EstimatorCreator
    {
    public:
      Creator() 
	: EstimatorCreator(), 
	  uh(NULL) 
      {}

      virtual ~Creator() {}

      inline void setSolution(DOFVector<double> *uh_)
      {
	uh = uh_;
      }

      /// Returns a new Estimator object.
      Estimator* create()
      {
	return new RecoveryEstimator(name, uh, row);
      }

    protected:
      DOFVector<double> *uh;
    };

    /// constructor
    RecoveryEstimator(std::string name, DOFVector<double> *uh, int r = -1);

    /// destructor.
    virtual ~RecoveryEstimator() {}

    /// implements \ref Estimator::estimate().
    virtual double estimate(double timestep = 0.0);


    /// Sets uh.
    inline void setUh(DOFVector<double> *uh)
    {
      uh_ = uh;
    }

    /// Sets f.
    inline void setFct(AbstractFunction<double, WorldVector<double> > *fct)
    {
      f_vec = fct;
    }

    ///
    inline void setFct(AbstractFunction<double, double> *fct)
    {
      f_scal = fct;
    }

    /// Sets auxiliar vector.
    inline void setAuxVec(DOFVector<double> *uh)
    {
      aux_vec = uh;
    }


    /// Gets recovery gradient.
    inline DOFVector<WorldVector<double> >* getRecGrd()
    {
      return rec_grd;
    }

    /// Gets higher-order recovery solution.
    inline DOFVector<double>* getRecUh()
    {
      return rec_uh;
    }


  protected:   
    /// finite element solution
    DOFVector<double> *uh_;

    /// absolute or relative error?
    int relative_;

    /// constant for scaling the estimator
    double C;

    /// recovery method
    int method;

    /// Working finite element space
    const FiniteElemSpace *feSpace;

    /// Degree of corresponding basic functions
    int degree;

    /// Basis functions for recovery vector.
    const BasisFunction *rec_basFcts;

    /// Recovery gradient
    DOFVector<WorldVector<double> > *rec_grd;

    /// Higher-order recovery solution
    DOFVector<double> *rec_uh;

    /// Diffusion coefficient (for flux recovery)
    AbstractFunction<double, WorldVector<double> > *f_vec;
    AbstractFunction<double, double> *f_scal;

    /// auxiliar vector
    DOFVector<double> *aux_vec;

    /// Recovery structure.
    Recovery *rec_struct;
  };

}

#endif
